#ifndef _PROCESSINFO_H_
#define _PROCESSINFO_H_

typedef enum _eProtocol
{
	E_SHMEM_PROTOCOL = 1,
	E_TCPIP_PROTOCOL,
	E_END_PROTOCOL
} E_PROTOCOL;

#define MAXPATHLEN					256
#define MAX_NUM_SHMEM_PROCS			8
#define MAX_WORLD_PROCS				256
#define MAX_NUM_MACHINES			64
#define CMD_NUMPROCS_STR			"-np"
#define CMD_CONFIGFILE_STR			"-f"
#define CMD_SINGLE_PROC_STR			"-s"
#define CMD_KILL_SINGLE_PROC_STR	"-ks"
#define CMD_KILL_MPI_JOB_STR		"-kj"
#define CMD_NOWAIT_STR				"-nowait"
#define CMD_DEBUG_STR				"-debug"
#define CMD_PORTNUMBER_STR			"-p"
#define CMD_GLOBALRANK_STR			"-GlobalRank"
#define CMD_JOBINFO_STR				"-JobInfo"
#define CMD_WORLDMAP_STR			"-WorldMap"
#define CMD_PROCESS_HANDLES_STR		"-ProcessHandles"
#define CMD_TCPIPPROC_STR			"-TCPIPListenerNeeded"
#define CMD_TCPIPPROC_NOTIFY_STR	"-TCPIPProcess"
#define LOCALMACHINE_STR			"LocalMachine"

typedef struct
{
	char szExecutable[MAXPATHLEN];
	DWORD dwLocalBufferNumber;
	DWORD dwMachineNumber;
	DWORD dwGlobalRank;
	HANDLE hSharedMemoryBuffer;
} MACHINEPROCESSINFO;

typedef struct
{
	char szMachineName[MAXPATHLEN];
	DWORD dwNumProcesses;
	MACHINEPROCESSINFO Process[MAX_NUM_SHMEM_PROCS];
} MACHINEJOBINFO;

typedef struct
{
	HANDLE hBarrierSemaphore;
	HANDLE hBarrierEvent;
	HANDLE hGlobalTerminatorEvent;
	HANDLE hLocalProcessHandles;
	HANDLE hTCPIPProcBuffer;
	DWORD dwNumMachines;
	DWORD dwWorldSize;
	DWORD dwMyMachineID;
	WORD  wPortNumber;
	MACHINEJOBINFO Machine[MAX_NUM_MACHINES];
} JOBINFO;

typedef struct
{
	DWORD dwMachineNumber;
	DWORD dwLocalBufferNumber;
	E_PROTOCOL eProtocol;
} PROCESSMAP;

typedef struct
{
	DWORD dwNumProcs;
	/*  this must be the the max number of shared memory
		processes plus the max number of concurrent protocols
		supported other than shared memory (currently 1 for TCP) */
	PROCESSMAP Process[MAX_NUM_MACHINES + 1];
} WORLDPROCESSMAP;

typedef struct
{
	DWORD dwNumHandles;
	HANDLE hProc[MAX_NUM_SHMEM_PROCS];
	HANDLE hThread[MAX_NUM_SHMEM_PROCS];
} LOCALPROCESSHANDLES;

#endif
