#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include <winsock.h>
#include "service.h"
#include "processinfo.h"
#include "mpiserve.h"
#include "SocketReadAndWrite.h"
#include "MPIServePacket.h"
#include "packets.h"
#include "infoprint.h"
#include "mpirun.h"
#include "crtdbg.h"

#include "fggout.h" /* FGG */

DWORD MPIRunStartMPIJob(JOBINFO *MemJI, WORLDPROCESSMAP *MemWPM)
{
	DWORD dwStatus = ERROR_SUCCESS;
	BOOL fStatus = FALSE;
	HANDLE hLocalProcessHandles = NULL;
	LOCALPROCESSHANDLES *LocalProcessHandles = NULL;
	HANDLE hJobInfoFileMap = NULL;
	HANDLE hWorldProcessMapFileMap = NULL;
	WORLDPROCESSMAP *WorldProcessMap = NULL;
	JOBINFO *JobInfo = NULL;
	SECURITY_ATTRIBUTES sa;
	int i;

	char FggOutput[1024]; /* FGG */
	HANDLE m_hFile;		  /* End FGG */

	/* FGG */
	m_hFile=NULL;
	if (SaveOutputToFile(FggOutput)!=0)
	{
		SECURITY_ATTRIBUTES g_sa = {sizeof(SECURITY_ATTRIBUTES),NULL,TRUE};
		
		m_hFile = CreateFile(FggOutput, 
			GENERIC_WRITE,
			FILE_SHARE_WRITE, 
			&g_sa,//NULL,
			CREATE_ALWAYS, 
			FILE_ATTRIBUTE_NORMAL, 
			NULL);
	}
	/* End FGG */

	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;
	hLocalProcessHandles = CreateFileMapping((HANDLE) 0xFFFFFFFF,	/* create a new one */
											&sa,					/* null security attributes */
											PAGE_READWRITE | SEC_COMMIT,	/* read and write to the memory */
											0,						/* high order bits */
											sizeof(LOCALPROCESSHANDLES),	/* low order bits */
											NULL);					/* no name */
	_ASSERTE(hLocalProcessHandles);
	if (NULL == hLocalProcessHandles)
		goto reporterror;

	LocalProcessHandles = (LOCALPROCESSHANDLES *)
							MapViewOfFile(hLocalProcessHandles,	/* handle to file mapping */
										FILE_MAP_ALL_ACCESS,		/* all access */
										0,							/* high order bits */
										0,							/* low order bits */
										0);							/* map the whole file */
	_ASSERTE(LocalProcessHandles);
	if (NULL == LocalProcessHandles)
		goto reporterror;	

	hWorldProcessMapFileMap = CreateFileMapping((HANDLE) 0xFFFFFFFF,	/* create a new one */
												&sa,					/* null security attributes */
												PAGE_READWRITE,			/* read and write to the memory */
												0,						/* high order bits */
												sizeof(WORLDPROCESSMAP),/* low order bits */
												NULL);					/* no name */
	_ASSERTE(hWorldProcessMapFileMap);
	if (NULL == hWorldProcessMapFileMap)
		goto reporterror;

	WorldProcessMap = (WORLDPROCESSMAP *) MapViewOfFile(hWorldProcessMapFileMap,	/* handle to file mapping */
														FILE_MAP_ALL_ACCESS,		/* all access */
														0,							/* high order bits */
														0,							/* low order bits */
														0);							/* map the whole file */
	_ASSERTE(WorldProcessMap);
	if (NULL == WorldProcessMap)
		goto reporterror;

	hJobInfoFileMap = CreateFileMapping((HANDLE) 0xFFFFFFFF,	/* create a new one */
										&sa,					/* null security attributes */
										PAGE_READWRITE | SEC_COMMIT,			/* read and write to the memory */
										0,						/* high order bits */
										sizeof(JOBINFO),		/* low order bits */
										NULL);					/* no name */
	_ASSERTE(hJobInfoFileMap);
	if (NULL == hJobInfoFileMap)
		goto reporterror;

	JobInfo = (JOBINFO *) MapViewOfFile(hJobInfoFileMap,	/* handle to file mapping */
										FILE_MAP_ALL_ACCESS,		/* all access */
										0,							/* high order bits */
										0,							/* low order bits */
										0);							/* map the whole file */
	_ASSERTE(JobInfo);
	if (NULL == JobInfo)
		goto reporterror;

	/*
		now that I have shared memory to write all the job information to,
		I can fill out the job information stuff
	*/

	/* copy over everything sent to me into shared memory */
	memcpy(JobInfo,MemJI,sizeof(JOBINFO));
	memcpy(WorldProcessMap,MemWPM,sizeof(WORLDPROCESSMAP));
	JobInfo->hLocalProcessHandles = hLocalProcessHandles;
	if (1 < JobInfo->dwNumMachines)
	{
		dwStatus = CreateShmemBuf(&JobInfo->hTCPIPProcBuffer);
		_ASSERTE(JobInfo->hTCPIPProcBuffer);
	}
	_ASSERTE(ERROR_SUCCESS == dwStatus);
	dwStatus = RunLocalMPIJob(JobInfo,
							  hJobInfoFileMap,
							  WorldProcessMap,
							  hWorldProcessMapFileMap,
							  LocalProcessHandles,
							  hLocalProcessHandles, m_hFile);
	if (ERROR_SUCCESS == dwStatus)
	{			
		for (i=0; i < (int)LocalProcessHandles->dwNumHandles; i++)
		{
			dwStatus = ResumeThread(LocalProcessHandles->hThread[i]);
			if (0xFFFFFFFF == dwStatus)
				goto abort;
		}
	}
	else
		goto abort;

	dwStatus = WaitForMultipleObjects(LocalProcessHandles->dwNumHandles,
									  LocalProcessHandles->hProc,
									  TRUE,
									  INFINITE);
	/* FGG */
	if (m_hFile != NULL) CloseHandle(m_hFile);
	/* End FGG */

	if (WAIT_FAILED == dwStatus)
		goto reporterror;
	else
		dwStatus = ERROR_SUCCESS;

abort:
	if (ERROR_SUCCESS != dwStatus)
	{
		for (i=0;i < MAX_NUM_SHMEM_PROCS; i++)
		{
			BOOL fStatus;
			if (LocalProcessHandles->hProc[i])
			{
				fStatus = TerminateProcess(LocalProcessHandles->hProc[i],(UINT)0xFFFFFFFF);
				_ASSERTE(fStatus);
				if (FALSE == fStatus)
					printf("Cannot kill process with handle [%d]!\n",
							LocalProcessHandles->hProc[i]);
				if (fStatus)
				{
					LocalProcessHandles->hProc[i] = NULL;
					LocalProcessHandles->hThread[i] = NULL;
				}
			}
		}
	}
	for (i=0;i < (int)LocalProcessHandles->dwNumHandles;i++)
	{
		DWORD dwExitCode = 0xFFFFFFFF;
		if (LocalProcessHandles->hProc[i])
			GetExitCodeProcess(LocalProcessHandles->hProc[i],&dwExitCode);
		printf("Process [%d] exited with status [%d]\n",i,dwExitCode);
	}
	if (LocalProcessHandles)
	{
		FreeLocalProcessHandles(LocalProcessHandles);
		UnmapViewOfFile(LocalProcessHandles);
	}
	if (hLocalProcessHandles)
	{
		CloseHandle(hLocalProcessHandles);
		JobInfo->hLocalProcessHandles = NULL;
	}
	if (WorldProcessMap)
	{
		FreeWorldProcessMap(WorldProcessMap);
		UnmapViewOfFile(WorldProcessMap);
	}
	if (hWorldProcessMapFileMap)
		CloseHandle(hWorldProcessMapFileMap);
	if (JobInfo)
	{
		FreeJobInfo(JobInfo);
		UnmapViewOfFile(JobInfo);
	}
	if (hJobInfoFileMap)
		CloseHandle(hJobInfoFileMap);

reporterror:
	dwStatus = GetLastError();
	return dwStatus;
}