//  MODULE:   ServiceStopAndStart.c
//
//  PURPOSE:  Contains the functions to start and stop the service
//
//  FUNCTIONS:
//	VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv)
//	VOID ServiceStop()
//
//  COMMENTS:
//
//  AUTHOR: Shane Hebert
//


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include <winsock.h>
#include "service.h"
#include "processinfo.h"
#include "mpiserve.h"

static HANDLE  hServerStopEvent = NULL;

//
//  FUNCTION: ServiceStart(DWORD dwArgc,LPTSTR *lpszArgv)
//
//  PURPOSE: workhorse function for service
//
//  PARAMETERS:
//    dwArgc - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//		This function is the workhorse of the service.
//		It actually is what the service is all about.
//
VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv)
{
	LONG nStatus;

    ///////////////////////////////////////////////////
    //
    // Service initialization
    //
    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_START_PENDING, // service state
        NO_ERROR,              // exit code
        3000))                 // wait hint
        goto cleanup;
    
    // create the event object. The control handler function signals
    // this event when it receives the "stop" control code.
    //
    hServerStopEvent = CreateEvent(
        NULL,    // no security attributes
        TRUE,    // manual reset event
        FALSE,   // not-signalled
        NULL);   // no name

    if ( hServerStopEvent == NULL)
        goto cleanup;

	// report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
        SERVICE_RUNNING,       // service state
        NO_ERROR,              // exit code
        0))                    // wait hint
        goto cleanup;

	{
		HANDLE hWorker;
		HANDLE hArray[2];
		DWORD dwThreadID;

		hWorker = CreateThread(
						NULL,
						0,
						SocketSetupAndListen,
						NULL,
						0,
						&dwThreadID);
		hArray[0] = hServerStopEvent;
		hArray[1] = hWorker;
		nStatus = WaitForMultipleObjects(2,hArray,FALSE,INFINITE);
		if (WAIT_OBJECT_0 == nStatus)
		{
			TerminateThread(hWorker,0);
			goto cleanup;
		}
		else if ((WAIT_OBJECT_0+1) == nStatus)
		{
			BOOL bStatus;
			DWORD dwExitCode;
			bStatus = GetExitCodeThread(hWorker,&dwExitCode);
			if (bStatus)
				nStatus = dwExitCode;
			goto cleanup;
		}
		CloseHandle(hWorker);
	}

cleanup:
	if (ERROR_SUCCESS != nStatus)
	{
		TCHAR szMsg[512];
		GetLastErrorText(szMsg,512);
		AddToMessageLog(szMsg);
	}
	if (hServerStopEvent)
        CloseHandle(hServerStopEvent);
}

//
//  FUNCTION: ServiceStop()
//
//  PURPOSE: stops the service and cleans up
//
//  PARAMETERS:
//		none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
VOID ServiceStop()
{
    if ( hServerStopEvent )
        SetEvent(hServerStopEvent);
}
