#include <crtdbg.h>
#include "windows.h"
#include "dm.h"
#include "packets.h"
#include "queue.h"
#include "queuedata.h"
#include "stdio.h"
#include "tchar.h"

#if !defined(SHMEM_AND_TCP) && !defined(SHMEM_ONLY)
#error "This module is for shared memory only!"
#endif


DWORD WINAPI SHMEMSendThread(LPVOID lpV)
{
	DWORD dwStatus;
	BOOL fStatus;
	MPID_WINNT_SHMEM_BUF_T **ppBuffers;
	MPID_PKT_T *pTmpPkt;
	LONG nErr;
	LONG nPartner;
	DWORD dwShortPacketsSent = 0;
	DWORD dwShortSyncPacketsSent = 0;
	DWORD dwLongPacketsSent = 0;
	DWORD dwLongSyncPacketsSent = 0;
	DWORD dwSyncAckPacketsSent = 0;
	HANDLE Events[2];
	extern HANDLE hEndProcessEvent;
	extern QUEUETYPE Queue[NUMQUEUES/2];
	extern MPID_NumShmemProcs;
	
	_ASSERTE(lpV);

	ppBuffers = (MPID_WINNT_SHMEM_BUF_T **)lpV;
	Events[0] = hEndProcessEvent;
	Events[1] = Queue[SHMEMSENDQUEUE].hSomethingQueuedEvent;


	while (TRUE)
	{
		LONG nDest;
		MPID_WINNT_SHMEM_BUF_T *pShmemDestBuffer;
		char *pDest;
		char *pSrc;
		extern int MPID_MyWorldRank;

		dwStatus = WaitForMultipleObjects(2,Events,FALSE,INFINITE);
		if (WAIT_FAILED == dwStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}
		if (WAIT_OBJECT_0 == dwStatus)
			/* this is the correct termination */
			goto cleanup;

		/* now we have a packet to send */
		fStatus = Dequeue(SHMEMSENDQUEUE,&pTmpPkt,&nPartner,&nErr);	
		if (!fStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}

		nDest = nPartner;
		if (MPID_NumShmemProcs < nPartner)
		{
			/*
			This is to redirect the packet to the TCP proxy process
			it will figure out where it really needs to go.
			*/
			nDest = MPID_NumShmemProcs + 1;
		}
		pTmpPkt->head.from = MPID_MyWorldRank;
		pTmpPkt->head.to = nPartner; /* the real destination */
		pShmemDestBuffer = &(ppBuffers[nDest][MPID_MyWorldRank]);
		dwStatus = WaitForSingleObject(pShmemDestBuffer->hBufferEmptied,INFINITE);
		if (WAIT_FAILED == dwStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}
		/* the buffer is empty so get the mutex */
		dwStatus = WaitForSingleObject(pShmemDestBuffer->hMutex,INFINITE);
		if (WAIT_FAILED == dwStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}
		switch (pTmpPkt->head.mode)
		{
		case MPID_PKT_SHORT:
			pShmemDestBuffer->Packet.head = pTmpPkt->head;
			pDest = pShmemDestBuffer->Packet.short_pkt.buffer;
			pSrc = pTmpPkt->short_pkt.buffer;
			if(0 != pTmpPkt->head.len)
			{
				__MEMCPY(pDest,pSrc,pTmpPkt->head.len);
			}
			
			fStatus = SetEvent(pShmemDestBuffer->hBufferFilled);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			
			fStatus = ReleaseMutex(pShmemDestBuffer->hMutex);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			
			ResetPacketEvent(SHMEMSENDQUEUE);
			
			__FREE(pTmpPkt);
			dwShortPacketsSent++;
			break;
		case MPID_PKT_SHORT_SYNC:
 			pShmemDestBuffer->Packet.head = pTmpPkt->head;

			pShmemDestBuffer->Packet.short_sync_pkt.sync_id = pTmpPkt->short_sync_pkt.sync_id;
			
			pDest = pShmemDestBuffer->Packet.short_sync_pkt.buffer;
			pSrc = pTmpPkt->short_sync_pkt.buffer;
			if(0 != pTmpPkt->head.len)
			{
				__MEMCPY(pDest,pSrc,pTmpPkt->head.len);
			}
			fStatus = SetEvent(pShmemDestBuffer->hBufferFilled);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			
			fStatus = ReleaseMutex(pShmemDestBuffer->hMutex);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			
			ResetPacketEvent(SHMEMSENDQUEUE);
			__FREE(pTmpPkt);
			dwShortSyncPacketsSent++;
			break;
		case MPID_PKT_LONG:
		case MPID_PKT_LONG_SYNC:
			/*  This one is a little more complicated.  Send the first 
				packet and then break the data up into packet sized 
				chunks (dynamically creating other packets */
			{
				int dwByteCount;
				int dwBytesToSend;
				MPID_PKT_HEAD_T long_data_head;
				
				/* send the header packet */
				/* Just need to send control info */
				
				pShmemDestBuffer->Packet.head = pTmpPkt->head;
				
				if (MPID_PKT_LONG_SYNC == pTmpPkt->head.mode)
		  			pShmemDestBuffer->Packet.long_sync_pkt.sync_id = 
											pTmpPkt->long_sync_pkt.sync_id;

				/* Let receiver to receive this control part */
				
				fStatus = SetEvent(pShmemDestBuffer->hBufferFilled);
				if (!fStatus)
				{
					dwStatus = GetLastError();
					_ASSERTE(FALSE);
				}
				
				fStatus = ReleaseMutex(pShmemDestBuffer->hMutex);
				if (!fStatus)
				{
					dwStatus = GetLastError();
					_ASSERTE(FALSE);
				}

				ResetPacketEvent(SHMEMSENDQUEUE);
				dwShortPacketsSent++;

				/* break up the rest */
				dwByteCount = 0;
				dwBytesToSend = pTmpPkt->head.len;
				pDest = pShmemDestBuffer->Packet.long_data_pkt.buffer;
				pSrc = pTmpPkt->long_pkt.pBuffer;

				/* Prepare the head for long data packets */
				long_data_head = pTmpPkt->head;
				long_data_head.mode = MPID_PKT_LONG_DATA;

				while (0 != dwBytesToSend)
				{
					dwStatus = WaitForSingleObject(pShmemDestBuffer->hBufferEmptied,INFINITE);
					if (WAIT_FAILED == dwStatus)
					{
						dwStatus = GetLastError();
						_ASSERTE(FALSE);
					}
					/* the buffer is empty so get the mutex */
					dwStatus = WaitForSingleObject(pShmemDestBuffer->hMutex,INFINITE);
					if (WAIT_FAILED == dwStatus)
					{
						dwStatus = GetLastError();
						_ASSERTE(FALSE);
					}

					if((signed int)(dwBytesToSend - MPID_PKT_MAX_DATA_SIZE) <= 0)
					{
						dwByteCount = dwBytesToSend;
						dwBytesToSend = 0;
					}
					else 
					{
						dwByteCount = MPID_PKT_MAX_DATA_SIZE;
						dwBytesToSend -= MPID_PKT_MAX_DATA_SIZE;
					}

					/* The length of the current long data packet is */
					long_data_head.len = dwByteCount;
									
					/* Put the head into the destination shmem buffer */
					pShmemDestBuffer->Packet.head = long_data_head;

					__MEMCPY(pDest,pSrc,dwByteCount);
					pSrc += dwByteCount;

					fStatus = SetEvent(pShmemDestBuffer->hBufferFilled);
					if (!fStatus)
					{
						dwStatus = GetLastError();
						_ASSERTE(FALSE);
					}
					fStatus = ReleaseMutex(pShmemDestBuffer->hMutex);
					if (!fStatus)
					{
						dwStatus = GetLastError();
						_ASSERTE(FALSE);
					}

				}
			}
			if (MPID_PKT_LONG == pTmpPkt->head.mode)
			{
				__FREE(pTmpPkt->long_pkt.pBuffer);
			}   
			else if (MPID_PKT_LONG_SYNC == pTmpPkt->head.mode) 
			{
				__FREE(pTmpPkt->long_sync_pkt.pBuffer);
			}
			__FREE(pTmpPkt);
			dwLongPacketsSent++;
			break;
		case MPID_PKT_SYNC_ACK:
			/* copy MODE fields */
			pShmemDestBuffer->Packet.sync_ack_pkt.mode = pTmpPkt->sync_ack_pkt.mode;
			//pShmemDestBuffer->Packet.sync_ack_pkt.context_id = pTmpPkt->sync_ack_pkt.context_id;
			//pShmemDestBuffer->Packet.sync_ack_pkt.lrank = pTmpPkt->sync_ack_pkt.lrank;
			/* copy sinc_id */
			pShmemDestBuffer->Packet.sync_ack_pkt.sync_id = pTmpPkt->sync_ack_pkt.sync_id;
			/* this is the whole MPID_PKT_SYNC_ACK */

			fStatus = SetEvent(pShmemDestBuffer->hBufferFilled);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			fStatus = ReleaseMutex(pShmemDestBuffer->hMutex);
			if (!fStatus)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			//ResetPacketEvent(SENDQUEUE);
			__FREE(pTmpPkt);
			dwSyncAckPacketsSent++;
			break;
		}

	}
cleanup:
	OutputDebugString("The SHMEMSENDThread terminates.");
	return dwStatus;
}

